from web3 import Web3, HTTPProvider
import json
from hashchain.ethereum.contract_interface import ABI, BYTECODE


class EthContract():
    """
    Solidity contract encapsulated in a Python object
    """
    def __init__(self):
        """
        Get the contract definition from'
        """
        self.abi = json.loads(ABI)
        self.bytecode = BYTECODE

    def deploy(self,
               sender_public_key: str,
               sender_private_key: str,
               provider_url: str) -> str:
        """
        Deploy the Solidity Smart contract to the Ethereum blockchain
        :param sender_public_key: public key of the sender
        :param sender_private_key: private key of the sender
        :param provider_url: address of the Ethereum connection provider
        :return: str: transaction hash
        """
        self.owner = Web3.toChecksumAddress(sender_public_key)
        self._owner_private_key = sender_private_key
        self._provider_address = provider_url
        self.w3 = Web3(HTTPProvider(self._provider_address))
        self.contract = self.w3.eth.contract(bytecode=self.bytecode,
                                             abi=self.abi)

        txn_dict = dict(
            nonce=self.w3.eth.getTransactionCount(self.owner),
            gasPrice=self.w3.eth.gasPrice,
            gas=2000000)

        txn = self.contract.constructor().buildTransaction(txn_dict)
        signed_txn = self.w3.eth.account.signTransaction(txn, self._owner_private_key)

        self.txn_hash = self.w3.eth.sendRawTransaction(signed_txn.rawTransaction)

        return self.txn_hash

    def get_txn_receipt(self) -> dict:
        """
        Wait until the transaction receipt is available and add the address variable to EthContract and returns the transaction receipt of the contract creation transaction.
        :return: dict: transaction receipt
        """
        txn_receipt = self.w3.eth.waitForTransactionReceipt(self.txn_hash)

        if txn_receipt:
            self.address = txn_receipt['contractAddress']

        return txn_receipt



import hashlib
import json


class Record():
    def __init__(self, content: dict, previous_hash: str = None):
        if not previous_hash:
            genesis = hashlib.sha3_256(b'0x0000000000000000000000000000000000000000000000000000000000000000')
            previous_hash = genesis.hexdigest()

        self.__content = json.dumps(content).encode('utf-8')
        self.__previous_hash = previous_hash
        self.__hash = hashlib.sha3_256()
        self.__hash.update(self.__content)
        self.__hash.update(previous_hash.encode('utf-8'))

    def __eq__(self, other: 'Record') -> bool:
        if self.get_hash() == other.get_hash():
            return True
        else:
            return False

    def get_hash(self) -> str:
        """
        Get the hex hash of the record
        :return: hex string
        """
        return self.__hash.hexdigest()

    def hex(self) -> str:
        """
        Get the hex hash of the record
        :return: hex string
        """
        return self.__hash.hexdigest()

    def get_content(self) -> dict:
        """
        Get the original content of the record
        :return: dict
        """
        return json.loads(self.__content.decode('utf-8'))

    def get_previous_hash(self) -> str:
        """
        Get the previous hash
        :return: hex string
        """
        return self.__previous_hash

    def update(self, new_content: dict):
        """
        Updates the record and recalculated the hash
        :param new_content: new record's content
        :return: None
        """
        self.__init__(new_content, self.__previous_hash)

    def to_dict(self) -> dict:
        """
        Return a dict of the complete record along with the hex string of the record's hash and the previous hash
        :return: dict
        """
        dict = self.get_content()
        dict['hash'] = self.get_hash()
        dict['previous_hash'] = self.get_previous_hash()
        return dict

    def to_json(self) -> str:
        """
        Returns the JSON of the complete record along with the hex string of the record's hash and the previous hash
        :return: JSON string
        """
        return json.dumps(self.to_dict())


class Chain():
    def __init__(self, content_dicts: list, last_hash: str = None):
        if last_hash:
            self.last_hash = last_hash
        else:
            self.last_hash = hashlib.sha3_256(
                b'0x0000000000000000000000000000000000000000000000000000000000000000').hexdigest()
        self.records = []
        for element in content_dicts:
            record = Record(element, self.last_hash)
            self.records.append(record)
            self.last_hash = record.get_hash()


def verify(records_dicts: list) -> bool:
    """
    Verifies a given list of records dicts
    :param records_dicts: list of Records objects
    :return: returns True if the list is valid. Raise ValueError if not valid.
    """
    for index, record in enumerate(records_dicts):
        content = {x : record[x] for x in record if x not in ['previous_hash','hash']}
        test_record = Record(content, record['previous_hash'])

        if record['hash'] != test_record.get_hash():
            raise ValueError(f"The record: {record} do no correspond to the hash provided :{record['hash']}")

        elif index >= 1 and test_record.get_previous_hash() != records_dicts[index - 1]['hash']:
            raise ValueError(f"The previous hash: {test_record.get_previous_hash()} do no correspond to the hash of the previous element: {records_dicts[index - 1]['hash']}")

    return True



from web3 import Web3, HTTPProvider
import hashlib


class EthConnector():
    """
    Connector to interact with a smart contract on the Ethereum blockchain.
    """
    def __init__(self,
                 contract_abi: str,
                 contract_address: str,
                 sender_public_key: str,
                 sender_private_key: str,
                 provider_url: str):
        """

        :param contract_abi: smart contract's abi JSON string
        :param contract_address: smart contract address
        :param sender_public_key: public key of the sender
        :param sender_private_key: private key of the sender
        :param provider_url: address of the Ethereum connection provider
        """
        self._contract_abi = contract_abi
        self._contract_address = Web3.toChecksumAddress(contract_address)
        self._sender_public_key = Web3.toChecksumAddress(sender_public_key)
        self._sender_private_key = sender_private_key
        self._provider_address = provider_url
        self.w3 = Web3(HTTPProvider(self._provider_address))
        self.contract = self.w3.eth.contract(address=contract_address, abi=contract_abi)

    def record(self, key: str, hash: str, wait = False):
        """
        Records the key and hash in the smart contract storage
        :param key: indexed key, used to retrieve the hash
        :param hash: hash of the record
        :param wait: wait for the transaction to receipt before completing
        :return: transaction hash
        """
        key_hash = hashlib.sha3_256(key.encode('utf-8'))
        hash_bytes = bytes.fromhex(hash)
        txn_dict = dict(
            nonce=self.w3.eth.getTransactionCount(self._sender_public_key),
            gasPrice=self.w3.eth.gasPrice,
            gas=2000000
        )

        txn = self.contract.functions.record(key_hash.digest(), hash_bytes).buildTransaction(txn_dict)
        signed_txn = self.w3.eth.account.signTransaction(txn, self._sender_private_key)
        txn_hash = self.w3.eth.sendRawTransaction(signed_txn.rawTransaction)

        if wait:
            self.w3.eth.waitForTransactionReceipt(txn_hash)
            return txn_hash
        else:
            return txn_hash

    def get_record(self, key: str):
        """
        Get the record hash from the smart contract storage
        :param key: unique key
        :return: hash
        """
        key_hash = hashlib.sha3_256(key.encode('utf-8'))
        return self.contract.functions.getHash(key_hash.digest()).call().hex()



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='hashchain',
    version='0.6.1',
    description='Helper to certify database entries, using hashchains qnd blockchains',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/svandegar/hashchain',
    author='Seraphin Vandegar',
    author_email='svandegar@hotmail.com',
    license='MIT',
    packages=['hashchain','hashchain.ethereum'],
    zip_safe=False,
    download_url='',
    install_requires=[
        'web3',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7'
    ]
)
